from .constants import PYPI_MIRRORS as PYPI_MIRRORS
from .hindsight import BackgroundObject as BackgroundObject
from .lazylib import meta as meta
from .molecules import (
    import_with_lock as import_with_lock,
    list_paginate as list_paginate,
    exec_new_win as exec_new_win,
    open_new_win as open_new_win,
    system_open as system_open,
    perform_pip_command as perform_pip_command,
    perform_pip_install as perform_pip_install,
    rrggbb_bg2fg as rrggbb_bg2fg
)
from .project import (
    find_python as find_python,
    distributions as distributions,
    getdist as getdist,
    create as create,
    get_builtin_plugins as get_builtin_plugins,
    find_env_file as find_env_file,
    recursive_find_env_config as recursive_find_env_config,
    recursive_update_env_config as recursive_update_env_config,
    get_toml_config as get_toml_config
)


import os
from pathlib import Path
import shlex
from shutil import which
import subprocess
from tempfile import mkstemp
from threading import Lock
from types import ModuleType
from typing import List, Literal, Optional, Tuple, TypeVar, Union, overload


from nonebot_desktop_wing.constants import LINUX_TERMINALS, WINDOWS

_import_lock = Lock()
T = TypeVar("T")


def import_with_lock(
    name: str,
    package: Optional[str] = None
) -> ModuleType:
    """Like `importlib.import_module(...)`, but using a lock to avoid
    conflicts when importing package.
    """
    from importlib import import_module
    with _import_lock:
        return import_module(name, package)


def list_paginate(lst: List[T], sz: int) -> List[List[T]]:
    """
    Cut a list to lists whose length are equal-to-or-less-than a specified
    size.

    - lst: `List[T]`    - a list to be cut.
    - sz: `int`         - max length for cut lists.

    - return: `List[List[T]]`
    """
    return [lst[st:st + sz] for st in range(0, len(lst), sz)]


def get_pause_cmd() -> str:
    """Get pause command on users' platforms."""
    if WINDOWS:
        return "pause"
    return "read -n1 -p 进程已结束，按任意键关闭。"


def get_terminal_starter() -> Tuple[str, ...]:
    """Get args for opening a new window, purposed for executing scripts."""
    if WINDOWS:
        return ("start", "cmd.exe", "/c")
    for te in LINUX_TERMINALS:
        if which(te) is not None:
            return (te, "-e")
    raise FileNotFoundError("no terminal emulator found")


def get_terminal_starter_pure() -> Tuple[str, ...]:
    """Get args for opening a new window, only for opening a new window."""
    if WINDOWS:
        return ("start", "cmd.exe")
    for te in LINUX_TERMINALS:
        if which(te) is not None:
            return (te,)
    raise FileNotFoundError("no terminal emulator found")


def gen_run_script(
    cmd: str, cwd: Union[str, Path, None] = None, activate_venv: bool = False
) -> str:
    """
    Generate executable scripts, for running commands in a new window.

    - cmd: `str`                    - commands to be executed.
    - cwd: `Union[str, Path, None]` - the first work directory to be set.
    - activate_venv: `bool`         - whether to attempt to find and activate
                                      the venv in `cwd`.

    - return: `str`                 - temp script path.
    """
    fd, fp = mkstemp(".bat" if WINDOWS else ".sh", "nbdesktop-")
    if not WINDOWS:
        os.chmod(fd, 0o755)
    with open(fd, "w") as f:
        if not WINDOWS:
            f.write(f"#!/usr/bin/env bash\n")

        if cwd is not None:
            pcwd = Path(cwd)
            if activate_venv and (pcwd / ".venv").exists():
                if WINDOWS:
                    f.write(
                        f"{pcwd / '.venv' / 'Scripts' / 'activate.bat'}\n"
                    )
                else:
                    f.write(f"source {pcwd / '.venv' / 'bin' / 'activate'}\n")

            f.write(f"cd \"{cwd}\"\n")
        f.write(f"{cmd}\n")
        f.write(f"{get_pause_cmd()}\n")
    return fp


def exec_nowin(
    cmd: str, cwd: Union[str, Path, None] = None,
    *, catch_output: bool = True
) -> Tuple[subprocess.Popen[bytes], str]:
    """
    Execute commands in a subprocess.
    
    - cmd: `str`                    - commands to be executed.
    - cwd: `Union[str, Path, None]` - the first work directory to be set.
    - catch_output: `bool`          - whether to catch output stdout and
                                      stderr.

    - return: `(Popen[bytes], str)` - the process running commands and temp
                                      script path.
    """
    sname = gen_run_script(cmd, cwd)
    return subprocess.Popen(
        shlex.join((*get_terminal_starter(), sname)), shell=True,
        stdout=subprocess.PIPE if catch_output else None,
        stderr=subprocess.STDOUT if catch_output else None
    ), sname


def exec_new_win(
    cmd: str, cwd: Union[str, Path, None] = None
) -> Tuple[subprocess.Popen[bytes], str]:
    """
    Execute commands in a new window.
    
    - cmd: `str`                    - commands to be executed.
    - cwd: `Union[str, Path, None]` - the first work directory to be set.

    - return: `(Popen[bytes], str)` - the process running commands and temp
                                      script path.
    """
    sname = gen_run_script(cmd, cwd)
    return subprocess.Popen(
        shlex.join((*get_terminal_starter(), sname)), shell=True,
    ), sname


def open_new_win(
    cwd: Union[str, Path, None] = None
) -> subprocess.Popen[bytes]:
    """
    Open a new terminal window.

    - cwd: `Union[str, Path, None]` - the first work directory to be set.

    - return: `Popen[bytes]`        - the process running new window.
    """
    return subprocess.Popen(
        shlex.join(get_terminal_starter_pure()), shell=True, cwd=cwd
    )


def system_open(
    fp: Union[str, Path], *, catch_output: bool = False
) -> subprocess.Popen[bytes]:
    """
    Use system applications to open a file path or URI.

    - fp: `Union[str, Path]`    - a file path or URI.
    - catch_output: `bool`      - whether to catch output stdout and stderr.

    - return: `Popen[bytes]`    - the process running external applications.
    """
    return subprocess.Popen(
        shlex.join(("start" if WINDOWS else "xdg-open", str(fp))), shell=True,
        stdout=subprocess.PIPE if catch_output else None,
        stderr=subprocess.STDOUT if catch_output else None
    )


@overload
def perform_pip_command(
    pyexec: str, command: str, *args: str,
    new_win: Literal[False] = False, catch_output: bool = False
) -> subprocess.Popen[bytes]:
    ...


@overload
def perform_pip_command(
    pyexec: str, command: str, *args: str, new_win: Literal[True] = True
) -> Tuple[subprocess.Popen[bytes], str]:
    ...


def perform_pip_command(
    pyexec: str, command: str, *args: str,
    new_win: bool = False, catch_output: bool = False
) -> Union[subprocess.Popen[bytes], Tuple[subprocess.Popen[bytes], str]]:
    """
    Run pip commands.

    - pyexec: `str`         - path to python executable.
    - command: `str`        - pip command.
    - *args: `str`          - args after pip command.
    - new_win: `bool`       - whether to open a new terminal window.
    - catch_output: `bool`  - whether to catch output stdout and stderr.

    - return:               - the process running commands (and temp script
                              path if `new_win` is set `True`).
        - `Popen[bytes] if new_win == False`
        - `(Popen[bytes], str) if new_win == True`
    """
    cmd = [pyexec, "-m", "pip", command, *args]
    if not new_win:
        return subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE if catch_output else None,
            stderr=subprocess.STDOUT if catch_output else None
        )
    return exec_new_win(shlex.join(cmd))


@overload
def perform_pip_install(
    pyexec: str, *packages: str, update: bool = False, index: str = "",
    new_win: Literal[False] = False, catch_output: bool = False
) -> subprocess.Popen[bytes]:
    ...


@overload
def perform_pip_install(
    pyexec: str, *packages: str, update: bool = False, index: str = "",
    new_win: Literal[True] = True, catch_output: bool = False
) -> Tuple[subprocess.Popen[bytes], str]:
    ...


def perform_pip_install(
    pyexec: str, *packages: str, update: bool = False, index: str = "",
    new_win: bool = False, catch_output: bool = False
) -> Union[subprocess.Popen[bytes], Tuple[subprocess.Popen[bytes], str]]:
    """
    Run pip install.

    - pyexec: `str`         - path to python executable.
    - *packages: `str`      - packages to be installed.
    - index: `str`          - index for downloading.
    - new_win: `bool`       - whether to open a new terminal window.
    - catch_output: `bool`  - whether to catch output stdout and stderr.

    - return:               - the process running commands (and temp script
                              path if `new_win` is set `True`).
        - `Popen[bytes] if new_win == False`
        - `(Popen[bytes], str) if new_win == True`
    """
    args = (*packages,)
    if update:
        args += ("-U",)
    if index:
        args += ("-i", index)
    return perform_pip_command(
        pyexec, "install", *args,
        new_win=new_win,  # type: ignore
        catch_output=catch_output
    )


def rrggbb_bg2fg(color: str) -> Literal['#000000', '#ffffff']:
    """
    Convert hex color code background to black or white.

    - color: `str`  - color code with the shape of '#rrggbb'

    - return: `str` - converted color code (`'#000000'` or `'#ffffff'`)
    """
    c_int = int(color[1:], base=16)
    # Formula for choosing color:
    # 0.2126 × R + 0.7152 × G + 0.0722 × B > 0.5
    #   => bright color ==> use opposite dark
    c_bgr: List[int] = []
    for _ in range(3):
        c_bgr.append(c_int & 0xff)
        c_int >>= 8
    b, g, r = (x / 255 for x in c_bgr)
    return (
        "#000000" if 0.2126 * r + 0.7152 * g + 0.0722 * b > 0.5 else "#ffffff"
    )


from threading import Thread
from typing import IO, Callable, Generic, List, Optional, ParamSpec, TypeVar

T = TypeVar("T")
AnyStr_T = TypeVar("AnyStr_T", str, bytes)
P = ParamSpec("P")


class BackgroundObject(Generic[P, T]):
    """A descriptor for running functions in background."""
    def __init__(
        self, func: Callable[P, T], *args: P.args, **kwargs: P.kwargs
    ) -> None:
        """
        Initialize (start) a call in background.

        - func: `(P...) -> T`   - A callable to be called in background.
        - *args: `P.args`       - Positional args for the callable.
        - **kwargs: `P.kwargs`  - Keyword args for the callable.
        """
        self._func = func
        self._thread = Thread(None, self._work, None, args, kwargs)
        self._thread.start()
        # bgobject_log_func(
        #     f"[BackgroundObject] '{func.__module__}.{func.__name__}' is "
        #     f"running in background with {args=}, {kwargs=}"
        # )

    def __get__(self, obj, objtype=None) -> T:
        return self.get()

    def _work(self, *args: P.args, **kwargs: P.kwargs) -> None:
        self._value = self._func(*args, **kwargs)
        # bgobject_log_func(
        #     f"[BackgroundObject] '{self._func.__module__}."
        #     f"{self._func.__name__}' is done with {args=}, {kwargs=}"
        # )

    def get(self) -> T:
        """Wait for value."""
        self._thread.join()
        return self._value


class RealtimeIOPipe(Generic[AnyStr_T]):
    def __init__(
        self, stream: IO[AnyStr_T], writer: Callable[[AnyStr_T], object]
    ) -> None:
        self.stream = stream
        self.thread = Thread(target=self._pull)
        self.writer = writer
        self.thread.start()

    def _pull(self) -> None:
        while True:
            try:
                self.writer(self.stream.read(1))
            except ValueError:
                break


from __future__ import annotations
import asyncio
from glob import glob
from pathlib import Path
import sys
from typing import (
    TYPE_CHECKING, Iterable, List, Literal, Optional, Tuple, Union, overload
)

from dotenv.main import DotEnv

from nonebot_desktop_wing.constants import WINDOWS
from nonebot_desktop_wing.lazylib import nb_cli
from nonebot_desktop_wing.molecules import perform_pip_install

if TYPE_CHECKING:
    from importlib.metadata import Distribution
    from nb_cli.config import Driver, Adapter, ConfigManager
    from subprocess import Popen


def find_python(fp: Union[str, Path]) -> str:
    """Find a python executable in a directory."""
    pfp = Path(fp)
    veexec = (
        pfp / ".venv"
        / ("Scripts" if WINDOWS else "bin")
        / ("python.exe" if WINDOWS else "python")
    )
    return str(veexec) if veexec.exists() else sys.executable


def distributions(*fp: str) -> Iterable[Distribution]:
    from importlib import metadata
    if fp:
        return metadata.distributions(path=list(fp))
    return metadata.distributions()


def getdist(root: Union[str, Path]) -> Iterable[Distribution]:
    """Get packages installed in a directory."""
    return (
        distributions(
            *(
                str(Path(root) / si)
                for si in
                glob(".venv/**/site-packages", root_dir=root, recursive=True)
            )
        )
    )


@overload
def create(
    fp: str,
    drivers: List[Driver],
    adapters: List[Adapter],
    dev: bool,
    usevenv: bool,
    index: Optional[str] = None,
    new_win: Literal[False] = False,
    catch_output: bool = False
) -> Popen[bytes]:
    ...


@overload
def create(
    fp: str,
    drivers: List[Driver],
    adapters: List[Adapter],
    dev: bool,
    usevenv: bool,
    index: Optional[str] = None,
    new_win: Literal[True] = True
) -> Tuple[Popen[bytes], str]:
    ...


def create(
    fp: str,
    drivers: List[Driver],
    adapters: List[Adapter],
    dev: bool,
    usevenv: bool,
    index: Optional[str] = None,
    new_win: bool = False,
    catch_output: bool = False
) -> Union[Popen[bytes], Tuple[Popen[bytes], str]]:
    """
    Create a new NoneBot project.

    - fp: `str`                 - path to target project (must be empty or not
                                  exist)
    - drivers: `List[Driver]`   - drivers to be installed.
    - adapters: `List[Adapter]` - adapters to be installed.
    - dev: `bool`               - whether to use a profile for developing
                                  plugins.
    - usevenv: `bool`           - whether to create a virtual environment.
    - index: `Optional[str]`    - index url for downloading packages.
    - new_win: `bool`           - whether to create a new window.
    - catch_output: `bool`      - whether to catch output stdout and stderr.

    - return:                   - the process running commands (and temp
                                  script path if `new_win` is set `True`).
        - `Popen[bytes] if new_win == False`
        - `(Popen[bytes], str) if new_win == True`
    """
    p = Path(fp)
    if p.exists():
        p.rmdir()
    nb_cli.handlers.create_project(
        "simple" if dev else "bootstrap",
        {
            "nonebot": {
                "project_name": p.name,
                "drivers": [d.dict() for d in drivers],
                "adapters": [a.dict() for a in adapters],
                "use_src": True
            }
        },
        str(p.parent)
    )
    dri_real = [d.project_link for d in drivers]
    adp_real = [a.project_link for a in adapters]
    dir_name = p.name.replace(" ", "-")
    venv_dir = p / ".venv"

    if usevenv:
        from venv import create as create_venv
        create_venv(venv_dir, prompt=dir_name, with_pip=True)

    pyexec = find_python(p)

    return perform_pip_install(
        pyexec, "nonebot2", *dri_real, *adp_real,
        index=index or "",
        new_win=new_win,  # type: ignore
        catch_output=catch_output
    )


def get_builtin_plugins(pypath: str) -> List[str]:
    """Get built-in plugins, using python in an environment."""
    return asyncio.run(
        nb_cli.handlers.list_builtin_plugins(python_path=pypath)
    )


def find_env_file(fp: Union[str, Path]) -> List[str]:
    """Find all dotenv files in a directory."""
    return glob(".env*", root_dir=fp)


def get_env_config(ep: Union[str, Path], config: str) -> Optional[str]:
    return DotEnv(ep).get(config)


def recursive_find_env_config(
    fp: Union[str, Path], config: str
) -> Optional[str]:
    """
    Recursively find a config in dotenv files.

    - fp: `Union[str, Path]`    - project directory.
    - config: `str`             - config string.

    - return: `Optional[str]`   - value of the config.
    """
    pfp = Path(fp)
    cp = pfp / ".env"
    if not cp.is_file():
        # Default profile is 'prod' if main profile does not exist.
        return get_env_config(pfp / ".env.prod", config)
    # Use main profile.
    glb = DotEnv(cp).dict()
    if config in glb:
        # Config in main profile.
        return glb[config]
    env = glb.get("ENVIRONMENT", None)  # Get specified profile.
    return env and get_env_config(pfp / f".env.{env}", config)


def recursive_update_env_config(
    fp: Union[str, Path], config: str, value: str
) -> None:
    """
    Recursively edit a config in dotenv files.

    - fp: `Union[str, Path]`    - project directory.
    - config: `str`             - config string.
    - value: `str`              - new value of the config.
    """
    pfp = Path(fp)
    cp = pfp / ".env"
    if not cp.is_file():
        # Default profile is 'prod' if main profile does not exist.
        cp = pfp / ".env.prod"
        useenv = DotEnv(cp).dict()
    else:
        # Use main profile.
        useenv = DotEnv(cp).dict()
        if config not in useenv:
            env = get_env_config(cp, "ENVIRONMENT")
            if env:
                # Specified profile is usable.
                cenv = DotEnv(pfp / f".env.{env}").dict()
                if config in cenv:
                    cp = pfp / f".env.{env}"
                    useenv = cenv

    useenv.update({config: value})

    with open(cp, "w") as f:
        f.writelines(f"{k}={v}\n" for k, v in useenv.items() if k and v)


def get_toml_config(basedir: Union[str, Path]) -> ConfigManager:
    """Get project TOML manager for a project."""
    basepath = Path(basedir)
    return nb_cli.config.ConfigManager(find_python(basepath), basepath / "pyproject.toml")


import asyncio
from nonebot_desktop_wing.hindsight import BackgroundObject
from nonebot_desktop_wing.molecules import import_with_lock
from nonebot_desktop_wing.resources import load_module_data_raw


class _nb_cli:
    handlers = BackgroundObject(import_with_lock, "nb_cli.handlers", "*")
    config = BackgroundObject(import_with_lock, "nb_cli.config", "*")


nb_cli = _nb_cli


class _meta:
    drivers = BackgroundObject(asyncio.run, nb_cli().handlers.load_module_data("driver"))
    adapters = BackgroundObject(asyncio.run, nb_cli().handlers.load_module_data("adapter"))
    plugins = BackgroundObject(asyncio.run, nb_cli().handlers.load_module_data("plugin"))
    raw_drivers = BackgroundObject(load_module_data_raw, "drivers")
    raw_adapters = BackgroundObject(load_module_data_raw, "adapters")
    raw_plugins = BackgroundObject(load_module_data_raw, "plugins")


meta = _meta